import torch
import datetime
import os
import copy
import shutil
import pickle
import gzip
import json

from metient import metient as met
from metient.util import vertex_labeling_util as vutil
from metient.util import data_extraction_util as dutil
from metient.util import eval_util as eutil
from metient.util import plotting_util as putil
from metient.lib import polytomy_resolver as prutil
from metient.lib import v_optimizer as voptim
from metient.lib import u_optimizer as uoptim

from metient.util.globals import *

torch.set_printoptions(precision=2)

if torch.cuda.is_available():
    torch.set_default_tensor_type(torch.cuda.FloatTensor)

def prune_histories(solutions):
    '''
    Only keep the unique, Pareto front of trees
    '''

    # Collect each unique solution's parsimony metrics
    all_pars_metrics = []
    unique_labelings = set()
    final_solutions = []
    for soln in solutions:
        tree = vutil.LabeledTree(soln.T, soln.V)
        if tree not in unique_labelings:
            final_solutions.append(soln)
            unique_labelings.add(tree)
            all_pars_metrics.append((soln.m, soln.c, soln.s))
    # Keep the pareto front of trees
    pareto_metrics, pruned_solutions = vutil.pareto_front(final_solutions, all_pars_metrics)
    print("pareto_metrics", set(pareto_metrics))

    return pruned_solutions

def rank_solutions(solution_set, print_config, needs_pruning=True):
    '''
    Return the sorted, top k solutions
    '''        
    # 1. Only keep unique, Pareto optimal histories
    if needs_pruning:
        final_solutions = prune_histories(solution_set)
    else:
        final_solutions = solution_set

    # 2. Sort the solutions from lowest to highest loss
    final_solutions = sorted(list(final_solutions))

    # 3. Return the k best solutions
    k = print_config.k_best_trees if len(final_solutions) >= print_config.k_best_trees else len(final_solutions)
    final_solutions = final_solutions[:k]
    return final_solutions

def create_solution_set(best_Vs, best_soft_Vs, best_Ts, G, O, p, node_idx_to_label, weights):
    # Make a solution set
    losses, (ms,cs,ss) = vutil.clone_tree_labeling_objective(vutil.to_tensor(best_Vs), vutil.to_tensor(best_soft_Vs), vutil.to_tensor(best_Ts), torch.stack([G for _ in range(len(best_Vs))]), O, p, weights, True)
    solution_set = []
    # Did we find a solution with primary-only single source seeding?
    has_pss_solution = False
    for loss,m,c,s,V,soft_V,T in zip(losses,ms,cs,ss,best_Vs,best_soft_Vs,best_Ts):
        if s == 1:
            has_pss_solution = True
        soln = vutil.VertexLabelingSolution(loss, m, c, s, V, soft_V, T, G, node_idx_to_label)
        solution_set.append(soln)
    return solution_set, has_pss_solution

def get_best_final_solutions(results, G, O, p, weights, print_config, 
                             node_idx_to_label, num_internal_nodes, needs_pruning):
    '''
    Prune unecessary poly nodes (if they weren't used) and return the top k solutions
    '''

    multiresult_has_pss_solution = False
    full_solution_set = []
    unique_labelings = set()

    for result in results:
        
        best_Vs, best_soft_Vs, best_Ts, poly_res = result
        solution_set, has_pss_solution = create_solution_set(best_Vs, best_soft_Vs, best_Ts, G, O, p, node_idx_to_label, weights)
        multiresult_has_pss_solution = multiresult_has_pss_solution or has_pss_solution

        # 1. Make solutions unique before we do all this additional post-processing work which is time intensive
        unique_solution_set = []
        for soln in solution_set:
            tree = vutil.LabeledTree(soln.T, soln.V)
            if tree not in unique_labelings:
                unique_solution_set.append(soln)
                unique_labelings.add(tree)
        # Don't need to recover a primary-only single source solution if we have already found one
        if not multiresult_has_pss_solution:
            unique_solution_set = recover_prim_ss_solutions(unique_solution_set, unique_labelings, weights, O, p)
            multiresult_has_pss_solution = True
        
        #  Remove any extra resolver nodes that don't actually help
        unique_solution_set = prutil.remove_extra_resolver_nodes(unique_solution_set, num_internal_nodes, poly_res, weights, O, p)
        full_solution_set.extend(unique_solution_set)

    return rank_solutions(full_solution_set, print_config, needs_pruning=needs_pruning)

def recover_prim_ss_solutions(solution_set, unique_labelings, weights, O, p):
    '''
    In hard (i.e. usually large input) cases where we are unable to find a 
    primary-only seeding solution, see if we can recover one by post-processing
    final solutions and removing any met-to-met migration edges, and add these
    to our final solution set
    '''
    
    expanded_solution_set = []
    for solution in solution_set:
        expanded_solution_set.append(solution)
        clusters = putil.seeding_clusters(solution.V,solution.T)
        new_V = copy.deepcopy(solution.V)
        for s in clusters:
            new_V[:,s] = p.T 
        loss, (m,c,s) = vutil.clone_tree_labeling_objective(new_V, solution.soft_V, solution.T, solution.G, O, p, weights, True)
        new_solution = vutil.VertexLabelingSolution(loss, m, c, s, new_V, solution.soft_V, solution.T, solution.G, solution.idx_to_label)
        unique_labeled_tree = vutil.LabeledTree(solution.T, new_V)
        if unique_labeled_tree not in unique_labelings:
            expanded_solution_set.append(new_solution)
            unique_labelings.add(unique_labeled_tree)
    return expanded_solution_set

def prep_inputs(tree_fns, tsv_fns, run_names, estimate_observed_clones, output_dir):

    if not (len(tree_fns) == len(tsv_fns) == len(run_names)):
        raise ValueError("Inputs Ts, tsv_fns, and run_names must have equal length (length = number of patients in cohort")

    if isinstance(tree_fns[0], str):
        Ts = []
        for tree_fn in tree_fns:
            Ts.append(dutil.get_adjacency_matrix_from_txt_edge_list(tree_fn))
    else:
        Ts = tree_fns
    
    # If we're not estimating the observed clones, the tsv being inputted doesn't need to be pooled
    if estimate_observed_clones:
        pooled_tsv_fns = []
        for tsv_fn, run_name in zip(tsv_fns, run_names):
            pooled_tsv_fns.append(dutil.pool_input_tsv(tsv_fn, output_dir, f"tmp_{run_name}"))
    else:
        for tsv_fn in tsv_fns:
            dutil.validate_prepooled_tsv(tsv_fn)
        pooled_tsv_fns = tsv_fns
    
    return Ts, pooled_tsv_fns

def evaluate_label_clone_tree(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
             O, batch_size, custom_colors, bias_weights, solve_polytomies):
    '''
    Observed clone proportions are inputted (in tsv_fns), only labeling of clone tree is needed
    '''
    return evaluate(tree_fn, tsv_fn, weights, print_config, output_dir, run_name,
                    O, batch_size, custom_colors, bias_weights, solve_polytomies, estimate_observed_clones=False)

def evaluate(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
             O, batch_size, custom_colors, bias_weights, solve_polytomies, estimate_observed_clones=True):
    
    Ts, pooled_tsv_fns = prep_inputs([tree_fn], [tsv_fn], [run_name], estimate_observed_clones, output_dir)
    assert isinstance(weights.mig, (float, int)), "Weights must be either a float or an int in evaluate mode"
    assert isinstance(weights.seed_site, (float, int)), "Weights must be either a float or an int in evaluate mode"

    T, pooled_tsv_fn = Ts[0], pooled_tsv_fns[0]  
    print(pooled_tsv_fn)  
    primary_sites = dutil.get_primary_sites(pooled_tsv_fn)
    if len(primary_sites) > 1:
        print("Multiple primaries given. Running each as primary")

    for primary_site in primary_sites:
        infer_migration_history(T, pooled_tsv_fn, primary_site, weights, print_config, output_dir, f"{run_name}_{primary_site}", 
                              O=O, batch_size=batch_size, custom_colors=custom_colors, bias_weights=bias_weights, 
                              mode="evaluate", solve_polytomies=solve_polytomies, estimate_observed_clones=estimate_observed_clones)
    if estimate_observed_clones:
        os.remove(pooled_tsv_fn) # cleanup pooled tsv

def calibrate_label_clone_tree(tree_fns, tsv_fns, print_config, output_dir, run_names,
                               Os, batch_size, custom_colors, bias_weights, solve_polytomies):
    '''
    Observed clone proportions are inputted (in tsv_fns), only labeling of clone tree is needed
    '''
    return calibrate(tree_fns, tsv_fns, print_config, output_dir, run_names,
                    Os, batch_size, custom_colors, bias_weights, solve_polytomies, estimate_observed_clones=False)

def calibrate(tree_fns, tsv_fns, print_config, output_dir, run_names,
              Os, batch_size, custom_colors, bias_weights, solve_polytomies,
              estimate_observed_clones=True):
    '''
    Estimate observed clone proportions and labeling of clone tree for a cohort of patients,
    calibrate parsimony weights to metastasis priors
    '''
    
    Ts, pooled_tsv_fns = prep_inputs(tree_fns, tsv_fns, run_names, estimate_observed_clones, output_dir)

    # Only use maximum parsimony metrics when initially searching for high likelihood trees
    weights = met.Weights(mig=DEFAULT_CALIBRATE_MIG_WEIGHTS, comig=DEFAULT_CALIBRATE_COMIG_WEIGHTS, 
                          seed_site=DEFAULT_CALIBRATE_SEED_WEIGHTS, gen_dist=0.0, 
                          organotrop=0.0)
    # Don't spend time making visualizations for calibrated trees
    visualize = print_config.visualize
    input_k = print_config.k_best_trees
    print_config.visualize = False

    calibrate_dir = os.path.join(output_dir, "calibrate")

    print(f"Saving results to {calibrate_dir}")

    if os.path.exists(calibrate_dir):
        shutil.rmtree(calibrate_dir)
        print(f"Overwriting existing directory at {calibrate_dir}")
    
    os.makedirs(calibrate_dir)

    # 1. Go through each patient and get migration history in calibrate mode
    for i in range(len(Ts)):
        print("\n*** Calibrating for patient:", run_names[i], "***")
        O = Os[i] if Os != None else None

        primary_sites = dutil.get_primary_sites(pooled_tsv_fns[i])
        if len(primary_sites) > 1:
            print("Multiple primaries given. Running each as primary")

        for primary_site in primary_sites:
            infer_migration_history(Ts[i], pooled_tsv_fns[i], primary_site, weights, print_config, calibrate_dir, f"{run_names[i]}_{primary_site}", 
                                    O=O, batch_size=batch_size, custom_colors=custom_colors, bias_weights=bias_weights,
                                    mode="calibrate", solve_polytomies=solve_polytomies, estimate_observed_clones=estimate_observed_clones)

    # 2. Find the best theta for this cohort
    best_theta = eutil.get_max_cross_ent_thetas(pickle_file_dirs=[calibrate_dir])
    rounded_best_theta = [round(v,3) for v in best_theta]
    with open(os.path.join(calibrate_dir, "best_theta.json"), 'w') as json_file:
        json.dump(rounded_best_theta, json_file, indent=2)

    ordered_sites = dutil.extract_ordered_sites(pooled_tsv_fns)
        
    # 3. Recalibrate trees using the best thetas
    print_config.visualize = visualize
    print_config.k_best_trees = input_k
    cal_weights = met.Weights(mig=[best_theta[0]*50], comig=best_theta[1]*50, seed_site=[best_theta[2]*50],
                          gen_dist=1.0, organotrop=1.0)
    
    # 4. Use the saved trees to rescore trees, visualize, and re-save 
    for i in range(len(Ts)):
        O = Os[i] if Os != None else None
        primary_sites = dutil.get_primary_sites(pooled_tsv_fns[i])
        for primary_site in primary_sites:
            run_name = f"{run_names[i]}_{primary_site}"
            with gzip.open(os.path.join(calibrate_dir, f"{run_name}.pkl.gz"), 'rb') as f:
                pckl = pickle.load(f)

            saved_U = torch.tensor(pckl[OUT_OBSERVED_CLONES_KEY])
            primary_sites = dutil.get_primary_sites(pooled_tsv_fns[i])
            primary_idx = ordered_sites[i].index(primary_site)
            p = torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(ordered_sites[i])).T

            reranked_solutions = rank_solutions(vutil.create_reweighted_solution_set_from_pckl(pckl, O, p, cal_weights),
                                                print_config, needs_pruning=False)
            
            putil.save_best_trees(reranked_solutions, saved_U, O, cal_weights, ordered_sites[i], print_config, 
                                  custom_colors, primary_site, calibrate_dir, run_name)
    
    if estimate_observed_clones:
        for pooled_tsv_fn in pooled_tsv_fns:
            os.remove(pooled_tsv_fn) # cleanup pooled tsv

    return best_theta

def validate_inputs(T, node_idx_to_label, ref, var, primary_site, ordered_sites, weights, O, mode):
    if not (T.shape[0] == T.shape[1]):
        raise ValueError(f"Number of tree nodes should be consistent (T.shape[0] == T.shape[1])")
    if (T.shape[0] != len(node_idx_to_label)):
        raise ValueError(f"Number of node_idx_to_label needs to equal shape of adjacency matrix.")
    if not torch.is_tensor(T):
        raise ValueError("T is not a PyTorch tensor.")
    if ref != None and var != None:
        if ref.shape != var.shape:
            raise ValueError(f"ref and var must have identical shape, got {ref.shape} and {var.shape}")
        if not (ref.shape[1] == var.shape[1] == T.shape[0]):
            raise ValueError(f"Number of mutations/mutation clusters should be consistent (ref.shape[1] == var.shape[1] == T.shape[0])")
        if not (ref.shape[0] == var.shape[0] == len(ordered_sites)):   
            raise ValueError(f"Length of ordered_sites should be equal to ref and var dim 0")
    if not vutil.is_tree(T):
        raise ValueError("Adjacency matrix T is empty or not a tree.")
    if not primary_site in ordered_sites:
        raise ValueError(f"{primary_site} not in ordered_sites: {ordered_sites}")
    if (weights.organotrop == 0.0 and O != None)  and mode == "evaluate":
        print(f"Warning: O matrix was given but organotropism parameter of weights is 0.")
    if (weights.organotrop != 0.0 and O == None):
        raise ValueError(f"O matrix was not given but organotropism parameter of weights is non-zero. Please pass an O matrix.")
    if mode != 'calibrate' and mode != 'evaluate':
        raise ValueError(f"Valid modes are 'evaluate' and 'calibrate'")
    for label in list(node_idx_to_label.values()):
        if ":" in label:
            raise ValueError(f"Unfortunately our visualization code uses pydot, which does not allow colons (:) in node names. Please use a different separator in 'character_label' values.")


def infer_migration_history(T, tsv_fn, primary_site, weights, print_config, output_dir, run_name, estimate_observed_clones=True,
                            O=None, lr=0.05, init_temp=20, final_temp=0.01,batch_size=-1, custom_colors=None, bias_weights=True,
                            mode="evaluate",solve_polytomies=False, needs_pruning=True):
    '''
    Args:
        T: numpy ndarray or torch tensor (shape: num_internal_nodes x num_internal_nodes). Adjacency matrix (directed) of the internal nodes.
        
        tsv_fn: path to tsv with the required columns: 
            ['anatomical_site_index', 'anatomical_site_label', 'cluster_index', 'character_label', 
            'ref', 'var', 'var_read_prob', 'site_category']

        weights: Weight object for how much to penalize each component of the loss
        
        print_config: PrintConfig object with options on how to visualize output
        
        output_dir: path for where to save output trees to

        run_name: e.g. patient name, used for naming output files.

    Optional:
        
        O: numpy ndarray or torch tensor (shape: 1 x  num_anatomical_sites).
        Matrix of organotropism values from primary tumor type to other sites (in order of anatomical
        site indices indicated in tsv_fn).

        bias_weights: whether to initialize weights higher to favor vertex labeling of primary for all internal nodes

        mode: can be "evaluate" or "calibrate"

    Returns:
        Corresponding info on the *best* tree:
        (1) edges of the tree (e.g. [('0', '1'), ('1', '2;3')])
        (2) vertex labeling as a dictionary (e.g. {'0': 'P', '1;3': 'M1'}),
        (3) edges for the migration graph (e.g. [('P', 'M1')])
        (4) dictionary w/ loss values for each component of the loss
        (5) how long (in seconds) the algorithm took to run
    '''

    start_time = datetime.datetime.now()

    # Extract inputs from tsv
    ref, var, omega, ordered_sites, node_idx_to_label, idx_to_observed_sites, G = dutil.extract_matrices_from_tsv(tsv_fn, estimate_observed_clones, T)
    
    # Validate inputs
    validate_inputs(T, node_idx_to_label, ref, var, primary_site, ordered_sites, weights, O, mode)
    print("ordered_sites",  ordered_sites)
    if batch_size == -1:
        batch_size = vutil.calculate_batch_size(T, ordered_sites, solve_polytomies)

    # When calibrating, we want to capture the Pareto front trees and not prune yet
    if mode == 'calibrate':
        print_config.k_best_trees = batch_size

    # Make the root index 0 (if it isn't already) to simplify indexing logic
    # Save the original root index to swap it back later though, since we don't 
    # want to confuse users with different cluster mappings etc.
    original_root_idx = vutil.get_root_index(T)
    T, ref, var, node_idx_to_label, G, idx_to_observed_sites, _, _ = vutil.restructure_matrices_root_index_zero(T, ref, var, node_idx_to_label, G, idx_to_observed_sites)
    assert(vutil.get_root_index(T) == 0)

    primary_idx = ordered_sites.index(primary_site)
    p = torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(ordered_sites)).T
    num_sites = len(ordered_sites)
    num_internal_nodes = T.shape[0]
    
    primary_site_label = ordered_sites[torch.nonzero(p)[0][0]]

    identical_clone_gen_dist = 0.0
    if G != None:
        identical_clone_gen_dist = torch.min(G[(G != 0)])/2.0

    B = vutil.mutation_matrix_with_normal_cells(T)

    # Keep a copy of input clone tree (T from now on has leaf nodes from U)
    input_T = copy.deepcopy(T)

    config = {
        "init_temp": init_temp,
        "final_temp": final_temp,
        "v_anneal_rate": 0.01,
        "t_anneal_rate": 0.01,
        "lr": lr,
        "first_max_iter": 50,
        "second_max_iter": 100,
        "first_v_interval": 15,
        "second_v_interval": 20,
        "batch_size": batch_size,
        "bias_weights": bias_weights,
        "solve_polytomies": solve_polytomies,
        # the genetic distance between two identical clones is a close to 0 but non-zero value
        "identical_clone_gen_dist": identical_clone_gen_dist,
        "num_v_optimization_runs": 4
    }

    ############ Step 1, optimize U ############

    u_optimizer = uoptim.ObservedClonesSolver(num_sites, num_internal_nodes, ref, var, omega, idx_to_observed_sites,
                                              B, input_T, G, node_idx_to_label, weights, config, estimate_observed_clones)
    u_result = u_optimizer.run()
    U, input_T, T, G, L, node_idx_to_label, num_internal_nodes, idx_to_observed_sites = u_result

    ############ Step 2, optimize V ############

    num_nodes_to_label = num_internal_nodes - 1 # we don't need to learn the root labeling

    v_optimizer = voptim.VertexLabelingSolver(L, T, p, G, O, weights, config, num_sites, num_nodes_to_label,
                                              node_idx_to_label, input_T, idx_to_observed_sites)
    results = v_optimizer.run()
    
    time_elapsed = (datetime.datetime.now() - start_time).total_seconds()
    if print_config.verbose:
        print(f"Time elapsed: {time_elapsed}")
        
    ############ Step 3, visualize and save outputs ############
    with torch.no_grad():
        
        final_solutions = get_best_final_solutions(results, v_optimizer.G, O, p, weights,
                                                   print_config, node_idx_to_label, num_internal_nodes, needs_pruning)

        print("# final solutions:", len(final_solutions))

        edges, vert_to_site_map, mig_graph_edges, loss_info = putil.save_best_trees(final_solutions, U, O, weights,
                                                                                    ordered_sites,print_config, custom_colors, 
                                                                                    primary_site_label, output_dir, run_name,
                                                                                    original_root_idx=original_root_idx)   

    return edges, vert_to_site_map, mig_graph_edges, loss_info, time_elapsed



import torch
from collections import OrderedDict
from metient.util import vertex_labeling_util as vutil
#from metient.lib.v_optimzer import VertexLabelingSolver

import numpy as np

class PolytomyResolver():

    def __init__(self, v_optimizer, nodes_w_polys, resolver_sites):
        '''
        This is post U matrix estimation, so T already has leaf nodes.
        '''
        
        # 1. Pad the adjacency matrix so that there's room for the new resolver nodes
        # nodes_w_polys are the nodes that have polytomies
        # (we place them in this order: given internal nodes, new resolver nodes, leaf nodes from U)
        T, G = v_optimizer.T, v_optimizer.G
        bs = v_optimizer.config['batch_size']

        num_new_nodes = 0
        for r in resolver_sites:
            num_new_nodes += len(r)

        num_leaves = v_optimizer.L.shape[1]
        num_internal_nodes = T.shape[0]-num_leaves
        T = torch.nn.functional.pad(T, pad=(0, num_new_nodes, 0, num_new_nodes), mode='constant', value=0)
        # 2. Shift T and G to make room for the new indices (so the order is input internal nodes, new poly nodes, leaves)
        idx1 = num_internal_nodes
        idx2 = num_internal_nodes+num_leaves
        T = torch.cat((T[:,:idx1], T[:,idx2:], T[:,idx1:idx2]), dim=1)
        if G != None:
            G = torch.nn.functional.pad(G, pad=(0, num_new_nodes, 0, num_new_nodes), mode='constant', value=0)
            G = torch.cat((G[:,:idx1], G[:,idx2:], G[:,idx1:idx2]), dim=1)

        # 3. Get each polytomy's children (these are the positions we have to relearn)
        children_of_polys = vutil.get_child_indices(T, nodes_w_polys)

        # 4. Initialize a matrix to learn the polytomy structure
        num_nodes_full_tree = T.shape[0]
        poly_adj_matrix = vutil.repeat_n(torch.ones((num_nodes_full_tree, len(children_of_polys)), dtype=torch.float32), bs)
        resolver_indices = [x for x in range(num_internal_nodes, num_internal_nodes+num_new_nodes)]

        nodes_w_polys_to_resolver_indices = OrderedDict()
        start_new_node_idx = resolver_indices[0]
        for parent_idx, r in zip(nodes_w_polys, resolver_sites):
            num_new_nodes_for_poly = len(r)
            if parent_idx not in nodes_w_polys_to_resolver_indices:
                nodes_w_polys_to_resolver_indices[parent_idx] = []

            for i in range(start_new_node_idx, start_new_node_idx+num_new_nodes_for_poly):
                nodes_w_polys_to_resolver_indices[parent_idx].append(i)
            start_new_node_idx += num_new_nodes_for_poly

        resolver_labeling = torch.zeros(v_optimizer.num_sites, len(resolver_indices))
        t = 0
        for sites in resolver_sites:
            for site in sites:
                resolver_labeling[site, t] = 1
                t += 1

        offset = 0
        for parent_idx in nodes_w_polys:
            child_indices = vutil.get_child_indices(T, [parent_idx])
            # make the children of polytomies start out as children of their og parent
            # with the option to "switch" to being the child of the new poly node
            poly_adj_matrix[:,parent_idx,offset:(offset+len(child_indices))] = 1.0
            # we only want to let these children choose between being the child
            # of their original parent or the child of this new poly node, which
            # we can do by setting all other indices to -inf
            mask = torch.ones(num_nodes_full_tree, dtype=torch.bool)
            new_nodes = nodes_w_polys_to_resolver_indices[parent_idx]
            mask_indices = new_nodes + [parent_idx]
            mask[[mask_indices]] = 0
            poly_adj_matrix[:,mask,offset:(offset+len(child_indices))] = float("-inf")
            offset += len(child_indices)

        poly_adj_matrix.requires_grad = True
        
        # 5. Initialize potential new nodes as children of the polytomy nodes
        for i in nodes_w_polys:
            for j in nodes_w_polys_to_resolver_indices[i]:
                T[i,j] = 1.0
                v_optimizer.node_idx_to_label[j] = [f"{i}pol{j}"]
                if G != None:
                    G[i,j] = v_optimizer.config['identical_clone_gen_dist']

        # 6. The genetic distance between a new node and its potential
        # new children which "switch" is the same distance between the new
        # node's parent and the child
        resolver_index_to_parent_idx = {}
        for poly_node in nodes_w_polys_to_resolver_indices:
            new_nodes = nodes_w_polys_to_resolver_indices[poly_node]
            for new_node_idx in new_nodes:
                resolver_index_to_parent_idx[new_node_idx] = poly_node

        if G != None:
            for new_node_idx in resolver_indices:
                parent_idx = resolver_index_to_parent_idx[new_node_idx]
                potential_child_indices = vutil.get_child_indices(T, [parent_idx])
                for child_idx in potential_child_indices:
                    G[new_node_idx, child_idx] = G[parent_idx, child_idx]

        v_optimizer.T = T
        v_optimizer.G = G
        self.latent_var = poly_adj_matrix
        self.nodes_w_polys = nodes_w_polys
        self.children_of_polys = children_of_polys
        self.resolver_indices = resolver_indices
        self.resolver_index_to_parent_idx = resolver_index_to_parent_idx
        self.resolver_labeling = resolver_labeling

def is_same_mig_hist_with_node_removed(poly_res, T, num_internal_nodes, V, remove_idx, children_of_removal_node, p, prev_m, prev_c, prev_s):
    '''
    Returns True if migration graph is the same or better after 
    removing node at index remove_idx
    '''

    '''
    If any of the following are true:
        (1) the polytomy resolver node is the same color as its parent,
        (2) the polytomy resolver node only has one child that is the same color as it,

        # (1) the polytomy resolver node is the same color as its parent and only has one child,
        # (2) the polytomy resolver node only has one child that is the same color as it,
        # (3) the polytomy resolver node is the same color as its parent and all of its children (that are internal nodes) are the same color
        # (4) the polytomy resolver node is the same color as its parent and only one of its children is a different color
    the migration history won't change by removing the polytomy resolver node. 
    
    If that is not true, check to see if the migration history changes by removing the node
    '''
    parent_idx = poly_res.resolver_index_to_parent_idx[remove_idx]
    # print(remove_idx, parent_idx, torch.argmax(V[:,parent_idx]).item(),torch.argmax(V[:,remove_idx]).item(), torch.argmax(V[:,parent_idx]).item()==torch.argmax(V[:,remove_idx]).item())
    remove_idx_color = torch.argmax(V[:,remove_idx]).item()
    is_same_color_as_parent = torch.argmax(V[:,parent_idx]).item() == remove_idx_color
    is_same_color_as_child = torch.argmax(V[:,children_of_removal_node[0]]).item() == remove_idx_color
    # Case 1
    if is_same_color_as_parent:
        return True
    
    # Case 2
    if len(children_of_removal_node)==1 and (is_same_color_as_child):
        return True
    
    # is_same_color_as_all_children = True
    # num_children_diff_color = 0
    # for child in children_of_removal_node:
    #     is_same_color_as_child = torch.argmax(V[:,child]).item() == remove_idx_color
    #     # Is an internal node child and is not same color
    #     if not is_same_color_as_child and child < num_internal_nodes:
    #         is_same_color_as_all_children = False
    #     if not is_same_color_as_child:
    #         num_children_diff_color += 1
    # # Case 3
    # if is_same_color_as_parent and is_same_color_as_all_children:
    #     return True
    # # Case 4
    # if is_same_color_as_parent and num_children_diff_color == 1:
    #     return True
    return False

    candidate_T, candidate_V = T.detach().clone(), V.detach().clone()
    # Attach children of the node to remove back to their original parent
    for child_idx in children_of_removal_node:
        candidate_T[parent_idx,child_idx] = 1.0
    candidate_T = np.delete(candidate_T, remove_idx, 0)
    candidate_T = np.delete(candidate_T, remove_idx, 1)
    candidate_V = np.delete(candidate_V, remove_idx, 1)
    new_m, new_c, new_s, _, _ = vutil.ancestral_labeling_metrics(vutil.add_batch_dim(candidate_V), candidate_T, None, None, p, True)
    # print(prev_m, prev_c, prev_s, new_m, new_c, new_s, ((prev_m >= int(new_m)) and (prev_c >= int(new_c)) and (prev_s >= int(new_s))))
    return ((prev_m >= int(new_m)) and (prev_c >= int(new_c)) and (prev_s >= int(new_s)))

def remove_nodes(removal_indices, V, T, G, node_idx_to_label):
    '''
    Remove polytomy resolver nodes from V, T, G and node_idx_to_label
    if they didn't actually help
    '''
    T = T.clone().detach()
    V = V.clone().detach()
    
    # Attach children of the node to remove to their original parent
    for remove_idx in removal_indices:
        parent_idx = np.where(T[:,remove_idx] > 0)[0][0]
        child_indices = vutil.get_child_indices(T, [remove_idx])
        for child_idx in child_indices:
            T[parent_idx,child_idx] = 1.0
    # Remove indices from T, V and G
    T = np.delete(T, removal_indices, 0)
    T = np.delete(T, removal_indices, 1)
    V = np.delete(V, removal_indices, 1)

    if G != None: 
        G = G.clone().detach()
        G = np.delete(G, removal_indices, 0)
        G = np.delete(G, removal_indices, 1)

    # Reindex the idx to label dict
    new_node_idx_to_label, _= vutil.reindex_dict(node_idx_to_label, removal_indices)
    
    return V, T, G, new_node_idx_to_label

def remove_extra_resolver_nodes(solution_set, num_internal_nodes, poly_res, weights, O, p):
    '''
    If there are any resolver nodes that were added to resolve polytomies but they 
    weren't used (i.e. 1. they have no children or 2. they don't change the 
    migration history), remove them
    '''

    if poly_res == None:
        return solution_set
    # prev_ms, prev_cs, prev_ss, _, _ = vutil.ancestral_labeling_metrics(vutil.to_tensor(best_Vs), vutil.to_tensor(best_Ts), None, None, p, True)
    out_solution_set = []
    for soln in solution_set:
        V, T, prev_m, prev_c, prev_s = soln.V, soln.T, soln.m, soln.c, soln.s
        nodes_to_remove = []
        for new_node_idx in poly_res.resolver_indices:
            children_of_new_node = vutil.get_child_indices(T, [new_node_idx])
            if len(children_of_new_node) == 0:
                nodes_to_remove.append(new_node_idx)
            elif is_same_mig_hist_with_node_removed(poly_res, T, num_internal_nodes,  V, new_node_idx, children_of_new_node, p, prev_m, prev_c, prev_s):
                nodes_to_remove.append(new_node_idx)

        if len(nodes_to_remove) != 0:
            new_V, new_T, new_G, new_node_idx_to_label = remove_nodes(nodes_to_remove, V, T, soln.G, soln.idx_to_label)
            loss, (new_m,new_c,new_s) = vutil.clone_tree_labeling_objective(new_V, soln.soft_V, new_T, new_G, O, p, weights, True)
            out_solution_set.append(vutil.VertexLabelingSolution(loss,new_m,new_c,new_s,new_V,soln.soft_V,new_T,new_G,new_node_idx_to_label))
        else:
            out_solution_set.append(soln)

    return out_solution_set


import numpy as np
import torch
import networkx as nx
import math
from queue import Queue
from metient.util.globals import *
from collections import deque
import copy

import scipy.sparse as sp
import pandas as pd
pd.options.display.float_format = '{:,.3f}'.format
pd.set_option('display.max_columns', None)

LAST_P = None

if torch.cuda.is_available():
    torch.set_default_tensor_type(torch.cuda.FloatTensor)

######################################################
##################### CLASSES ########################
######################################################

    
# Defines a unique adjacency matrix and vertex labeling
class LabeledTree:
    def __init__(self, tree, labeling):
        if (tree.shape[0] != tree.shape[1]):
            raise ValueError("Adjacency matrix should have shape (num_nodes x num_nodes)")
        if (tree.shape[0] != labeling.shape[1]):
            raise ValueError("Vertex labeling matrix should have shape (num_sites x num_nodes)")

        self.tree = tree
        self.labeling = labeling

    def _nonzero_tuple(self, tensor):
        # Get the indices of non-zero elements and convert to a hashable form
        indices = torch.nonzero(tensor, as_tuple=False)
        return tuple(map(tuple, indices.tolist()))

    def __hash__(self):
        # Compute a hash based on the positions of non-zero entries
        return hash((self._nonzero_tuple(self.labeling), self._nonzero_tuple(self.tree)))

    def __eq__(self, other):
        # Check for equality based on the positions of non-zero entries in both tensors
        if not isinstance(other, LabeledTree):
            return False
        return (self._nonzero_tuple(self.labeling) == self._nonzero_tuple(other.labeling) and
                self._nonzero_tuple(self.tree) == self._nonzero_tuple(other.tree))

    def __str__(self):
        A = str(np.where(self.tree == 1))
        V = str(np.where(self.labeling == 1))
        return f"Tree: {A}\nVertex Labeling: {V}"

def convert_pars_metric_to_int(x):
    if isinstance(x, int):
        return x
    elif isinstance(x, torch.Tensor) and x.numel() == 1:
        return int(x.item())  
    raise ValueError(f"Got unexpected value for parsimony metric {x}")

# Convenience object to package information needed for a final solution
class VertexLabelingSolution:
    def __init__(self, loss, m, c, s, V, soft_V, T, G, idx_to_label):
        self.loss = loss
        self.m = convert_pars_metric_to_int(m)
        self.c = convert_pars_metric_to_int(c)
        self.s = convert_pars_metric_to_int(s)
        self.V = V
        self.T = T
        self.G = G
        self.idx_to_label = idx_to_label
        self.soft_V = soft_V

    # Override the comparison operator
    def __lt__(self, other):
        return self.loss < other.loss
    
    def _nonzero_tuple(self, tensor):
        # Get the indices of non-zero elements and convert to a hashable form
        indices = torch.nonzero(tensor, as_tuple=False)
        return tuple(map(tuple, indices.tolist()))

    def __hash__(self):
        # Compute a hash based on the positions of non-zero entries
        return hash((self._nonzero_tuple(self.V), self._nonzero_tuple(self.T)))

    def __eq__(self, other):
        # Check for equality based on the positions of non-zero entries in both tensors
        if not isinstance(other, VertexLabelingSolution):
            return False
        return (self._nonzero_tuple(self.V) == self._nonzero_tuple(other.V) and
                self._nonzero_tuple(self.T) == self._nonzero_tuple(other.T))



# Convenience object to package information needed for known 
# labelings/indices of vertex labeling matrix V
class FixedVertexLabeling:
    def __init__(self, known_indices, unknown_indices, known_labelings):
        self.known_indices = known_indices
        self.unknown_indices = unknown_indices
        self.known_labelings = known_labelings

######################################################
############# CALCULATING PARSIMONY METRICS ##########
######################################################

def migration_number(site_adj):
    '''
    Args:
        - site_adj: batch_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j
    Returns:
        - migration number: number of total migrations between sites (no same site to same site migrations)
    '''
    site_adj_trace = torch.diagonal(site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    m = torch.sum(site_adj, dim=(1, 2)) - site_adj_trace
    return m

def seeding_site_number(site_adj_no_diag):
    '''
    Args:
        - site_adj_no_diag: batch_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j, and no same site migrations are included
    Returns:
        - seeding site number: number of sites that have outgoing edges
    '''
    row_sums_site_adj = torch.sum(site_adj_no_diag, axis=2)
    # can only have a max of 1 for each site (it's either a seeding site or it's not)
    binarized_row_sums_site_adj = torch.sigmoid(BINARY_ALPHA * (2*row_sums_site_adj - 1)) # sigmoid for soft thresholding
    s = torch.sum(binarized_row_sums_site_adj, dim=(1))
    return s

def comigration_number(site_adj, A, VA, VT, X, update_path_matrix):
    '''
    Args:
        - site_adj: batch_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j
        - A: Adjacency matrix (directed) of the full tree (batch_size x num_nodes x num_nodes)
        - VA: V*A
        - VT: transpose of V 
        - X: VT*V (1 if node i and node j are the same color)
        - update_path_matrix: whether we need to update the path matrix or we can use a cached version
        (need to update when we're actively resolving polytomies)
    Returns:
        - comigration number: a subset of the migration edges between two anatomical sites, such that 
        the migration edges occur on distinct branches of the clone tree. It is the number of multi-edges in 
        migration graph G (V*A*V.T)
    '''

    VAT = torch.transpose(VA, 2, 1)
    W = VAT @ VA # 1 if two nodes' parents are the same color
    Y = torch.sum(torch.mul(VAT, 1-VT), axis=2) # Y has a 1 for every node where its parent has a diff color
    shared_par_and_self_color = torch.mul(W, X) # 1 if two nodes' parents are same color AND nodes are same color
    # tells us if two nodes are (1) in the same site and (2) have parents in the same site
    # and (3) there's a path from node i to node j
    global LAST_P # this is expensive to compute, so hash it if we don't need to update it
    if LAST_P != None and not update_path_matrix:
        P = LAST_P
    else:
        P = path_matrix(A, remove_self_loops=True)
        LAST_P = P

    shared_path_and_par_and_self_color = torch.sum(torch.mul(P, shared_par_and_self_color), axis=2)
    repeated_temporal_migrations = torch.sum(torch.mul(shared_path_and_par_and_self_color, Y), axis=1)
    binarized_site_adj = torch.sigmoid(BINARY_ALPHA * (2 * site_adj - 1))
    bin_site_trace = torch.diagonal(binarized_site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    c = torch.sum(binarized_site_adj, dim=(1,2)) - bin_site_trace + repeated_temporal_migrations
    return c

def genetic_distance_score(G, m, A, X):
    '''
    Args:
        - G: Matrix of genetic distances between internal nodes (shape: batch_size x num_internal_nodes x num_internal_nodes).
             Lower values indicate lower branch lengths, i.e. more genetically similar.
        - m: vector of migration numbers (length = batch_size)
        - A: Adjacency matrix (directed) of the full tree (batch_size x num_nodes x num_nodes)
        - X: VT*V (1 if node i and node j are the same color)
    Returns:
        - genetic distance score, summed over batch_size # of solutions
    '''
    g = 0
    if G != None:
        # Calculate if 2 nodes are in diff sites and there's an edge between them (i.e there is a migration edge)
        R = torch.mul(A, (1-X))
        adjusted_G = -torch.log(G+0.01)
        R = torch.mul(R, adjusted_G)
        g = torch.sum(R, dim=(1,2))/(m)
    return g

def organotropism_score(O, site_adj_no_diag, p, bs, num_sites):
    '''
    Args:
        - O: Array of frequencies with which the primary cancer type seeds site i (shape: num_anatomical_sites).
        - site_adj_no_diag: batch_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j, and no self loops
        - p: one-hot vector indicating site of the primary
        - bs: batch_size (number of samples)
        - num_sites: number of anatomical sites
    Returns:
        - organotropism score, summed over batch_size # of solutions
    '''
    o = 0
    if O != None:
        # the organotropism frequencies can only be used on the first 
        # row, which is for the migrations from primary cancer site to
        # other metastatic sites (we don't have frequencies for every 
        # site to site migration)
        prim_site_idx = torch.nonzero(p)[0][0]
        O = O.repeat(bs,1).reshape(bs, O.shape[0])
        adjusted_freqs = -torch.log(O+0.01)
        num_mig_from_prim = site_adj_no_diag[:,prim_site_idx,:]
        organ_penalty = torch.mul(num_mig_from_prim, adjusted_freqs)
        o = torch.sum(organ_penalty, dim=(1))/torch.sum(num_mig_from_prim, dim=(1))
    return o

def ancestral_labeling_metrics(V, A, G, O, p, update_path_matrix):

    single_A = A
    bs = V.shape[0]
    num_sites = V.shape[1]
    A = A if len(A.shape) == 3 else repeat_n(single_A, bs) 

    # Compute matrices used for all parsimony metrics
    VA = V @ A
    VT = torch.transpose(V, 2, 1)
    site_adj = VA @ VT
    # Remove the same site transitions from the site adjacency matrix
    site_adj_no_diag = torch.mul(site_adj, repeat_n(1-torch.eye(num_sites, num_sites), bs))
    X = VT @ V # 1 if two nodes are the same color

    # 1. Migration number
    m = migration_number(site_adj)

    # 2. Seeding site number
    s = seeding_site_number(site_adj_no_diag)

    # 3. Comigration number
    c = comigration_number(site_adj, A, VA, VT, X, update_path_matrix)

    # 4. Genetic distance
    g = genetic_distance_score(G, m, A, X)

    # 5. Organotropism
    o = organotropism_score(O, site_adj_no_diag, p, bs, num_sites)

    return m, c, s, g, o

def calc_entropy(V, soft_V):
    '''
    Args:
        - V: one-hot vertex labeling matrix
        - soft_V: underlying theta parameters of V
    Returns:
        - entropy for the categorical variable representing each node's vertex label
    '''
    eps = 1e-7 # to avoid nans when values in soft_V get very close to 0
    return -torch.sum(torch.mul(soft_V, torch.log2(soft_V+eps)), dim=(1, 2)) / V.shape[2]

def get_repeating_weight_vector(bs, weight_list):
    # Calculate the number of times each weight should be repeated
    total_weights = len(weight_list)
    repeats = bs // total_weights
    remaining_elements = bs % total_weights

    # Create a list where each weight is repeated 'repeats' times
    repeated_list = [weight for weight in weight_list for _ in range(repeats)]

    # Add remaining elements to match the batch size exactly
    if remaining_elements > 0:
        additional_weights = weight_list[:remaining_elements]
        additional_repeated = [weight for weight in additional_weights]
        repeated_list += additional_repeated

    # Convert the list to a tensor
    weights_vec = torch.tensor(repeated_list)

    return weights_vec

def get_mig_weight_vector(bs, weights):
    return get_repeating_weight_vector(bs, weights.mig)

def get_seed_site_weight_vector(bs, weights):
    return get_repeating_weight_vector(bs, weights.seed_site)

def clone_tree_labeling_loss_with_computed_metrics(m, c, s, g, o, e, weights, bs=1):

    # Combine all 5 components with their weights
    # Explore different weightings
    if isinstance(weights.mig, list) and isinstance(weights.seed_site, list):
        mig_weights_vec = get_mig_weight_vector(bs, weights)
        seeding_sites_weights_vec = get_seed_site_weight_vector(bs, weights)
        mig_loss = torch.mul(mig_weights_vec, m)
        seeding_loss = torch.mul(seeding_sites_weights_vec, s)
        labeling_loss = (mig_loss + weights.comig*c + seeding_loss + weights.gen_dist*g + weights.organotrop*o+ weights.entropy*e)
        
    else:
        mig_loss = weights.mig*m
        seeding_loss = weights.seed_site*s
        labeling_loss = (mig_loss + weights.comig*c + seeding_loss + weights.gen_dist*g + weights.organotrop*o+ weights.entropy*e)
    return labeling_loss

def clone_tree_labeling_objective(V, soft_V, A, G, O, p, weights, update_path_matrix):
    '''
    Args:
        V: Vertex labeling of the full tree (batch_size x num_sites x num_nodes)
        A: Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
        G: Matrix of genetic distances between internal nodes (shape:  num_internal_nodes x num_internal_nodes).
        Lower values indicate lower branch lengths, i.e. more genetically similar.
        O: Array of frequencies with which the primary cancer type seeds site i (shape: num_anatomical_sites).
        p: one-hot vector indicating site of the primary
        weights: Weights object

    Returns:
        Loss to score the ancestral vertex labeling of the given tree. This combines (1) migration number, (2) seeding site
        number, (3) comigration number, and optionally (4) genetic distance and (5) organotropism.
    '''
    V = add_batch_dim(V)
    soft_V = add_batch_dim(soft_V)
    m, c, s, g, o = ancestral_labeling_metrics(V, A, G, O, p, update_path_matrix)
    # Entropy
    e = calc_entropy(V, soft_V)

    labeling_loss = clone_tree_labeling_loss_with_computed_metrics(m, c, s, g, o, e, weights, bs=V.shape[0])

    return labeling_loss, (m, c, s)

######################################################
######### POST U MATRIX ESTIMATION UTILITIES #########
######################################################

def get_leaf_labels_from_U(U):
    
    U = U[:,1:] # don't include column for normal cells
    internal_node_idx_to_sites = {}
    for node_idx in range(U.shape[1]):
        for site_idx in range(U.shape[0]):
            node_U = U[site_idx,node_idx]
            if node_U > U_CUTOFF:
                if node_idx not in internal_node_idx_to_sites:
                    internal_node_idx_to_sites[node_idx] = []
                internal_node_idx_to_sites[node_idx].append(site_idx)
    
    return internal_node_idx_to_sites


def full_adj_matrix_from_internal_node_idx_to_sites_present(input_T, input_G, idx_to_sites_present, num_sites, G_identical_clone_val):
    '''
    All non-zero values of U represent extant clones (leaf nodes of the full tree).
    For each of these non-zero values, we add an edge from parent clone to extant clone.
    '''
    num_leaves = sum(len(lst) for lst in idx_to_sites_present.values())
    full_adj = torch.nn.functional.pad(input=input_T, pad=(0, num_leaves, 0, num_leaves), mode='constant', value=0)
    leaf_idx = input_T.shape[0]
    # Also add branch lengths (genetic distances) for the edges we're adding.
    # Since all of these edges we're adding represent genetically identical clones,
    # we are going to add a very small but non-zero value.
    full_G = torch.nn.functional.pad(input=input_G, pad=(0, num_leaves, 0, num_leaves), mode='constant', value=0) if input_G is not None else None

    leaf_labels = []
    # Iterate through the internal nodes that we want to attach leaf nodes to
    for internal_node_idx in idx_to_sites_present:
        # Attach a leaf node for every site that this internal node is observed in
        for site in idx_to_sites_present[internal_node_idx]:
            full_adj[internal_node_idx, leaf_idx] = 1
            if input_G is not None:
                full_G[internal_node_idx, leaf_idx] = G_identical_clone_val
            leaf_idx += 1
            leaf_labels.append(site)
    # Anatomical site labels of the leaves
    L = torch.nn.functional.one_hot(torch.tensor(leaf_labels), num_classes=num_sites).T
    return full_adj, full_G, L
    
def full_adj_matrix_using_inputted_observed_clones(input_T, input_G, idx_to_sites_present, num_sites, G_identical_clone_val):
    '''
    Use inputted observed clones to fill out T and G by adding leaf nodes
    '''
    full_adj, full_G, L = full_adj_matrix_from_internal_node_idx_to_sites_present(input_T, input_G, idx_to_sites_present, num_sites, G_identical_clone_val)
    return L, full_adj, full_G

def full_adj_matrix_using_inferred_observed_clones(U, input_T, input_G, num_sites, G_identical_clone_val):
    '''
    Use inferred observed clones to fill out T and G by adding leaf nodes
    '''
    internal_node_idx_to_sites = get_leaf_labels_from_U(U)
    full_adj, full_G, L = full_adj_matrix_from_internal_node_idx_to_sites_present(input_T, input_G, internal_node_idx_to_sites, num_sites, G_identical_clone_val)
    
    return full_adj, full_G, L, internal_node_idx_to_sites

def remove_leaf_indices_not_observed_sites(removal_indices, U, input_T, T, G, node_idx_to_label, idx_to_observed_sites):
    '''
    Remove clone tree leaf nodes that are not detected in any sites. 
    These are not well estimated
    '''
    if len(removal_indices) == 0:
        return U, input_T, T, G, node_idx_to_label, idx_to_observed_sites
    
    for remove_idx in removal_indices:
        child_indices = get_child_indices(T, [remove_idx])
        assert len(child_indices) == 0

    # Remove indices from input_T, full T (now with observed clones), U and G
    T = np.delete(T, removal_indices, 0)
    T = np.delete(T, removal_indices, 1)

    # Remove indices from T, U and G
    input_T = np.delete(input_T, removal_indices, 0)
    input_T = np.delete(input_T, removal_indices, 1)

    U = np.delete(U, removal_indices, 1)

    if G != None: 
        G = np.delete(G, removal_indices, 0)
        G = np.delete(G, removal_indices, 1)

    # Reindex the idx to label dict
    new_node_idx_to_label, old_index_to_new_index = reindex_dict(node_idx_to_label, removal_indices)
    new_idx_to_observed_sites = {}
    for old_idx in idx_to_observed_sites:
        new_idx = old_index_to_new_index[old_idx]
        new_idx_to_observed_sites[new_idx] = idx_to_observed_sites[old_idx]

    return U, input_T, T, G, new_node_idx_to_label, new_idx_to_observed_sites

######################################################
################## RANDOM UTILITIES ##################
######################################################

def mutation_matrix_with_normal_cells(T):
    B = mutation_matrix(T)
    # Add a row of zeros to account for the non-cancerous root node
    B = torch.vstack([torch.zeros(B.shape[1]), B])
    # Add a column of ones to indicate that every clone is a descendent of the non-cancerous root node
    B = torch.hstack([torch.ones(B.shape[0]).reshape(-1,1), B])
    return B

def adjacency_matrix_to_edge_list(adj_matrix):
    edges = []
    for i, j in tree_iterator(adj_matrix):
        edges.append((i, j))
    return edges

def reindex_dict(original_dict, indices_to_remove):
    # Create a new dictionary to hold the re-indexed entries
    new_dict = {}
    
    old_index_to_new_index = {}
    # Initialize the new index
    new_index = 0
    # Iterate through the original dictionary in sorted index order
    for old_index in sorted(original_dict.keys()):
        # Skip the indices that need to be removed
        if old_index in indices_to_remove:
            continue
        
        # Assign the new index to the current label
        new_dict[new_index] = original_dict[old_index]
        old_index_to_new_index[old_index] = new_index
        # Increment the new index
        new_index += 1
    
    return new_dict, old_index_to_new_index

def add_batch_dim(x):
    if len(x.shape) == 3:
        return x
    return x.reshape(1, x.shape[0], x.shape[1])

def to_tensor(t):
    if torch.is_tensor(t):
        return t
    return torch.stack(t)

def remove_leaf_nodes_idx_to_label_dicts(dicts):
    '''
    After running migration history inference, the leaf nodes 
    (U nodes) get added to the idx to label dicts when plotting 
    and saving to pickle files, so we don't want to do that twice
    when in calibrate mode
    '''
    new_dicts = []
    for i,dct in enumerate(dicts):
        new_dicts.append(copy.deepcopy(dct))
        for key in dct:
            if dct[key][1] == True:
                del new_dicts[i][key]
            else:
                new_dicts[i][key] = new_dicts[i][key][0]
    return new_dicts

def create_reweighted_solution_set_from_pckl(pckl, O, p, weights):
    # Make a solution set from the pickled files
    Ts, Vs, soft_Vs, Gs = pckl[OUT_ADJ_KEY], pckl[OUT_LABElING_KEY], pckl[OUT_SOFTV_KEY], pckl[OUT_GEN_DIST_KEY]
    idx_to_label_dicts = remove_leaf_nodes_idx_to_label_dicts(pckl[OUT_IDX_LABEL_KEY])
    solution_set = []
    for T, V, soft_V, G, idx_to_label in zip(Ts, Vs, soft_Vs, Gs, idx_to_label_dicts):
        loss, (m,c,s) = clone_tree_labeling_objective(torch.tensor(V), torch.tensor(soft_V), torch.tensor(T), torch.tensor(G), O, p, weights, True)
        solution_set.append(VertexLabelingSolution(loss, m, c, s, torch.tensor(V), torch.tensor(soft_V), torch.tensor(T), torch.tensor(G), idx_to_label))
    return solution_set

def calculate_batch_size(T, sites, solve_polytomies):
    '''
    Calculate the number of samples to initialize for a run based on 
    the number of tree nodes, the number of anatomical sites, and if we're
    solving polytomies
    '''
    num_nodes = T.shape[0]
    num_sites = len(sites)
    min_size = 64
    min_size += num_nodes*num_sites*4

    if solve_polytomies:
        min_size *= 2

    # cap this to a reasonably high sample size
    min_size = min(min_size, 60000)
    print("calculate_batch_size", min_size)
    return min_size

def tree_iterator(T):
    ''' 
    Iterate an adjacency matrix, yielding i and j for all values = 1
    '''
    # Enumerating through a torch tensor is pretty computationally expensive,
    # so convert to a sparse matrix to efficiently access non-zero values
    T = T if isinstance(T, np.ndarray) else T.detach().numpy()
    T = sp.coo_matrix(T)
    for i, j in zip(T.row, T.col):
        yield i,j

def bfs_iterator(tree, start_node):
    '''
    Iterate an adjacency matrix in breadth first search order
    '''
    queue = deque([start_node])  # Start the queue with the given start node index
    visited = set(queue)         # Mark the start node as visited

    # BFS loop that yields nodes in the order they are visited
    while queue:
        current = queue.popleft()
        yield current  # Yield the current node

        # Check each adjacency in the row for the current node
        for neighbor, is_connected in enumerate(tree[current]):
            if is_connected and neighbor not in visited:
                visited.add(neighbor)
                queue.append(neighbor)

def has_leaf_node(adjacency_matrix, start_node, num_internal_nodes):
    num_nodes = adjacency_matrix.size(0)
    visited = torch.zeros(num_nodes, dtype=torch.bool)
    queue = Queue()

    queue.put(start_node)
    visited[start_node] = True

    while not queue.empty():
        current_node = int(queue.get())
        # Get neighbors of the current node
        neighbors = torch.nonzero(adjacency_matrix[current_node]).squeeze(1)

        # Enqueue unvisited neighbors
        for neighbor in neighbors:
            neighbor = int(neighbor)
            if not visited[neighbor]:
                queue.put(neighbor)
                visited[neighbor] = True
                # it's a leaf node if its not in the clone tree idx label dict
                if neighbor > num_internal_nodes:
                    return True
    return False

def get_root_index(T):
    '''
    Returns the root idx (node with no inbound edges) from adjacency matrix T
    '''

    candidates = set([x for x in range(len(T))])
    #print("candidates", candidates)
    for _, j in tree_iterator(T):
        if j not in candidates:
            print("T", adjacency_matrix_to_edge_list(T))
            print("j", j)
        candidates.remove(j)
    msg = "More than one" if len(candidates) > 1 else "No"
    assert (len(candidates) == 1), f"{msg} root node detected"

    return list(candidates)[0]

def reverse_bfs_order(A):
    '''
    Returns nodes in reverse bfs order of adjacency matrix A
    '''
    root_idx = get_root_index(A)
    nodes = []
    for x in bfs_iterator(A, root_idx):
        nodes.append(x)
    nodes.reverse()
    return nodes

def get_leaves(A):
    '''
    Returns leaves of adjacency matrix A
    '''
    return [int(x) for x in torch.where(A.sum(dim=1)==0)[0]]

def get_descendants(A, i):
    '''
    Returns all descendant nodes of node i in  adjacency matrix A
    '''
    path = path_matrix(A,remove_self_loops=True)
    descendants = [int(x) for x in torch.where(path[i]==1)[0]]
    return descendants

def swap_keys(d, key1, key2):
    if key1 in d and key2 in d:
        # Both keys exist, swap their values
        d[key1], d[key2] = d[key2], d[key1]
    elif key1 in d and key2 not in d:
        # Only key1 exists, move its value to key2
        d[key2] = d.pop(key1)
    elif key2 in d and key1 not in d:
        # Only key2 exists, move its value to key1
        d[key1] = d.pop(key2)
    # If neither key exists, do nothing
    return d

def restructure_matrices_root_index_zero(adj_matrix, ref_matrix, var_matrix, node_idx_to_label, gen_dist_matrix, idx_to_observed_sites):
    '''
    Restructure the inputs so that the node at index 0 becomes the root node.
    '''
    og_root_idx = get_root_index(adj_matrix)
    return restructure_matrices(og_root_idx, 0, adj_matrix, ref_matrix, var_matrix, 
                                node_idx_to_label, gen_dist_matrix, idx_to_observed_sites, None, None)

def restructure_matrices(source_root_idx, target_root_idx, adj_matrix, ref_matrix, var_matrix, 
                         node_idx_to_label, gen_dist_matrix, idx_to_observed_sites, V, U):
    '''
    Restructure the inputs so that the order of nodes has source_root_idx and target_root_idx swapped

    Returns:
        - Restructured adjacency matrix, reference matrix, variant matrix, node_idt_to_label dcitionary,
        and genetic distance matricx
    '''
    if source_root_idx == target_root_idx:
        # Nothing to restructure here!
        return adj_matrix, ref_matrix, var_matrix, node_idx_to_label, gen_dist_matrix, idx_to_observed_sites, V, U
    
    new_order = [x for x in range(len(adj_matrix))]
    new_order[source_root_idx] = target_root_idx
    new_order[target_root_idx] = source_root_idx
    # Swap rows and columns to move the first row and column to the desired position
    swapped_adjacency_matrix = adj_matrix[new_order, :][:, new_order]

    # Swap columns 
    swapped_ref_matrix = ref_matrix[:, new_order] if ref_matrix != None else None
    swapped_var_matrix = var_matrix[:, new_order] if var_matrix != None else None
    swapped_V = V[:, new_order] if V != None else None
    
    if U == None:
        swapped_U = None
    else:
        normal_cells = U[:,0].view(-1, 1)
        U_order = [x+1 for x in new_order if x < U.shape[1]-1]
        swapped_U = U[:, U_order]
        swapped_U = torch.cat((normal_cells, swapped_U), dim=1)
        
    if gen_dist_matrix == None:
        swapped_gen_dist_matrix = None
    else:
        swapped_gen_dist_matrix = gen_dist_matrix[new_order, :][:, new_order]
    
    if idx_to_observed_sites == None:
        idx_to_observed_sites = None
    else:
        idx_to_observed_sites = swap_keys(idx_to_observed_sites, source_root_idx, target_root_idx)

    node_idx_to_label = swap_keys(node_idx_to_label, source_root_idx, target_root_idx)
    
    return swapped_adjacency_matrix, swapped_ref_matrix, swapped_var_matrix, node_idx_to_label, swapped_gen_dist_matrix, idx_to_observed_sites, swapped_V, swapped_U

def nodes_w_leaf_nodes(adj_matrices, num_internal_nodes):
    '''
    Args:
        - adj_matrices: 3d matrix, where each inner matrix is a 2d adjacency matric
        - num_internal_nodes: number of nodes in the clone tree which are not leaf nodes
          indicating clone presences

    Returns:
        A list of lists, where list i is for adjacency matrix i, and the inner list is a list
        of booleans indicated whether node j has a leaf clone presence node or not
    '''
    # Convert the 3D adjacency matrix to a PyTorch tensor
    adj_tensor = torch.tensor(adj_matrices, dtype=torch.float32)
    
    # Get dimensions
    _, n, _ = adj_tensor.size()

    # Step 2: Create a mask for columns with indices >= num_internal_nodes
    col_indices = torch.arange(n)
    col_mask = col_indices >= num_internal_nodes

    # Step 3: Apply the column mask to the 3D adjacency matrix
    # Broadcast the col_mask to the shape of the last dimension of adj_tensor
    filtered_cols = adj_tensor[:, :, col_mask]

    # Step 4: Check for each row in each 2D matrix if there's at least one '1' in the filtered columns
    mask = filtered_cols.any(dim=2)

    return mask

def print_U(U, B, node_idx_to_label, ordered_sites, ref, var):
    cols = ["GL"]+[";".join([str(i)]+node_idx_to_label[i][:2]) for i in range(len(node_idx_to_label))]
    U_df = pd.DataFrame(U.detach().numpy(), index=ordered_sites, columns=cols)

    print("U\n", U_df)
    F_df = pd.DataFrame((var/(ref+var)).numpy(), index=ordered_sites, columns=cols[1:])
    print("F\n", F_df)
    Fhat_df = pd.DataFrame((U @ B).detach().numpy()[:,1:], index=ordered_sites, columns=cols[1:])
    print("F hat\n", Fhat_df)

def top_k_integers_by_count(lst, k, min_num_sites, cutoff):
    # find unique sites and their counts
    unique_sites, counts = np.unique(lst, return_counts=True)
    # filter for sites that occur at least min_num_sites times
    filtered_values = unique_sites[counts >= min_num_sites]
    # recount after filtering
    unique_sites, counts = np.unique(filtered_values, return_counts=True)
    # sort based on counts in descending order
    sorted_indices = np.argsort(counts)[::-1]
    # get the top k integers and their counts
    if len(unique_sites) > k and cutoff:
        return list(unique_sites[sorted_indices[:k]])
        
    return list(unique_sites)

def traverse_until_Uleaf_or_mult_children(node, input_T, internal_node_idx_to_sites):
    """
    Traverse down the tree on a linear branch until a leaf node is found, and then
    return that node
    
    Special case where a node isn't detected in any sites and is on a long linear branch,
    so there are no children or grandchildren observed to help bias towards
    """
    if node in internal_node_idx_to_sites:
        return node
    
    children = input_T[node].nonzero()[:,0].tolist()
    
    if len(children) == 1:
        return traverse_until_Uleaf_or_mult_children(children[0], input_T, internal_node_idx_to_sites)
    
    return None  # no node has a leaf or node has more than one child
   
def get_resolver_sites_of_node_and_children(input_T, node_idx_to_sites, num_children, node_idx, include_children, min_num_sites, cutoff):
    '''
    Looking at the sites that node_idx and node_idx's children are in,
    figure out the sites that the polytomy resolver nodes should be initialized with
    (any sites that are detected >= thres times)
    '''
   
    clone_tree_children = input_T[node_idx].nonzero()[:,0].tolist()
    # calculated based on clone tree children and leaf nodes estimated form U
    num_possible_resolvers = math.floor(num_children/2)
    # get the sites that the node and node's children are in
    if include_children:
        node_and_children = clone_tree_children+ [int(node_idx)]
    else:
        node_and_children = [int(node_idx)]
    sites = []
    for key in node_and_children:
        if key in node_idx_to_sites:
            sites.extend(node_idx_to_sites[key])
    # print("node_idx", node_idx, "node_and_children", node_and_children, "sites", sites)
    # Special case where the node of interest (node_idx) isn't detected in any sites
    # and is on a linear branch 
    if not include_children and len(sites) == 0 and len(clone_tree_children) == 1:
        leaf_on_linear_branch = traverse_until_Uleaf_or_mult_children(node_idx, input_T, node_idx_to_sites)
        # print("leaf_on_linear_branch", leaf_on_linear_branch)
        if leaf_on_linear_branch != None:
            sites = node_idx_to_sites[leaf_on_linear_branch]
    
    return top_k_integers_by_count(sites, num_possible_resolvers, min_num_sites, cutoff)
    
def get_k_or_more_children_nodes(input_T, T, internal_node_idx_to_sites, k, include_children, min_num_sites, cutoff=True):
    '''
    returns the indices and proposed labeling for nodes
    that are under nodes with k or more children
    e.g. 
    input_T =[[0,1,1,],
             [0,0,0],
             [0,0,0],]
    T = [[0,1,1,1,0,0,0],
         [0,0,0,0,1,0,1],
         [0,0,0,0,0,1,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0],]
    internal_node_idx_to_sites = {0:[1], 1:[0,1], 2:[1}
    k = 3
    include_children = True
    min_num_sites = 1
    returns ([0], [[1]]) (node at index 0 has 3 children, and the resolver node under it should be
    place in site 1, since node 0 and its child node 1 are both detected in site 1)

    if include_children is False, we only look at node 0's U values and not its children's U values
    '''
    row_sums = torch.sum(T, axis=1)
    node_indices = list(np.where(row_sums >= k)[0])
    filtered_node_indices = []
    all_resolver_sites = []
    for node_idx in node_indices:
        # get the sites that node_idx and node_idx's children are estimated in U
        resolver_sites = get_resolver_sites_of_node_and_children(input_T, internal_node_idx_to_sites, row_sums[node_idx], node_idx, include_children, min_num_sites, cutoff)
        # print("resolver_sites", resolver_sites)
        # a resolver node wouldn't help for this polytomy
        if len(resolver_sites) == 0:
            continue
        filtered_node_indices.append(node_idx)
        all_resolver_sites.append(resolver_sites)
    return filtered_node_indices, all_resolver_sites

def find_first_branching_point(adj_matrix):
    out_degrees = adj_matrix.sum(dim=1)  # Calculate out-degrees of each node
    for node, out_degree in enumerate(out_degrees):
        if out_degree > 1:
            return node
    return None  # Return None if no branching point is found

import torch
import networkx as nx
from itertools import product

# Helper function to build a NetworkX tree from an adjacency matrix
def build_tree(adj_matrix):
    G = nx.DiGraph()
    n = adj_matrix.shape[0]
    for i in range(n):
        for j in range(n):
            if adj_matrix[i, j] == 1:
                G.add_edge(i, j)
    return G

# Helper function to find the first common branching point in two graphs
def find_first_branching_point(adj_matrix):
    out_degrees = adj_matrix.sum(dim=1)  # Calculate out-degrees of each node
    for node, out_degree in enumerate(out_degrees):
        if out_degree > 1:
            return node
    return None  # Return None if no branching point is found

# Helper function to extract subtrees rooted at the given node
def extract_subtrees(G, root):
    subtrees = []
    for child in G.successors(root):
        subtree = nx.dfs_tree(G, source=child)
        subtrees.append(subtree)
    return subtrees

# Helper function to combine subtrees and labels from multiple solutions
def combine_subtrees(subtrees_list, labels_list, common_root):
    combinations = []
    num_subtrees = len(subtrees_list[0])
    
    # Generate all combinations of subtrees from each solution
    for indices in product(range(num_subtrees), repeat=len(subtrees_list)):
        combined_adj_matrix = torch.zeros_like(subtrees_list[0][0].adj_matrix)
        combined_labels = torch.zeros_like(labels_list[0])
        
        for i, idx in enumerate(indices):
            subtree = subtrees_list[i][idx]
            labels = labels_list[i]
            # Combine adjacency matrices
            for edge in subtree.edges:
                combined_adj_matrix[edge[0], edge[1]] = 1
            # Combine labels
            for node in subtree.nodes:
                combined_labels[:, node] = labels[:, node]
        
        combinations.append((combined_adj_matrix, combined_labels))
    return combinations

def get_child_indices(T, indices):
    '''
    returns the indices of direct children of the nodes at indices
    '''
    all_child_indices = []

    for parent_idx in indices:
        children = np.where(T[parent_idx,:] > 0)[0]
        for child_idx in children:
            if child_idx not in all_child_indices:
                all_child_indices.append(child_idx)

    return all_child_indices

def find_parents_children(T, node):
    num_nodes = len(T)
    
    # Find parents: Look in the node's column
    parents = [i for i in range(num_nodes) if T[i][node] != 0]
    
    # Find children: Look in the node's row
    children = [j for j in range(num_nodes) if T[node][j] != 0]
    
    return parents, children

def repeat_n(x, n):
    '''
    Repeats tensor x 'n' times along the first axis, returning a tensor
    w/ dim (n, x.shape[0], x.shape[1])
    '''
    if n == 0:
        return x
    return x.repeat(n,1).reshape(n, x.shape[0], x.shape[1])

def path_matrix(T, remove_self_loops=False):
    '''
    T is a numpy ndarray or tensor adjacency matrix (where Tij = 1 if there is a path from i to j)
    '''
    bs = 1 if len(T.shape) == 2 else T.shape[0]
    # Path matrix that tells us if path exists from node i to node j
    I = torch.eye(T.shape[1]).repeat(bs, 1, 1)  # Repeat identity matrix along batch dimension
    B = torch.logical_or(T, I).int()  # Convert to int for more efficient matrix multiplication
    # Initialize path matrix with direct connections
    P = B.clone()
    
    # Floyd-Warshall algorithm
    for k in range(T.shape[1]):
        # Compute shortest paths including node k
        B = torch.logical_or(B, B[:, :, k].unsqueeze(2) & B[:, k, :].unsqueeze(1))
        # Update path matrix
        P |= B
        
    if remove_self_loops:
        P = torch.logical_xor(P, I.int())
    return P.squeeze(0) if len(T.shape) == 2 else P

def mutation_matrix(A):
    '''
    A is an numpy ndarray or tensor adjacency matrix (where Aij = 1 if there is a path from i to j)

    returns a mutation matrix B, which is a subclone x mutation binary matrix, where Bij = 1
    if subclone i has mutation j.
    '''
    return path_matrix(A.T, remove_self_loops=False)

def get_adj_matrix_from_edge_list(edge_list):
    T = []
    nodes = set([node for edge in edge_list for node in edge])
    T = [[0 for _ in range(len(nodes))] for _ in range(len(nodes))]
    for edge in edge_list:
        T[ord(edge[0]) - 65][ord(edge[1]) - 65] = 1
    return torch.tensor(T, dtype = torch.float32)

def is_tree(adj_matrix):
    rows, cols = np.where(adj_matrix == 1)
    edges = zip(rows.tolist(), cols.tolist())
    g = nx.Graph()
    g.add_edges_from(edges)
    return (not nx.is_empty(g) and nx.is_tree(g))

def pareto_front(solutions, all_pars_metrics):
    """ 
    Args:
        - solutions: list of VertexLabelingSolutions of length n
        - all_pars_metrics: list of tuples with parsimony mentrics
        for (migration #, comigration #, seeding site #) of length n

    Returns:
        - a list of Pareto optimal parsimony metrics and VertexLabelingSolutions
    """
    
    # Start with an empty Pareto front
    pareto_front = []

    # Loop through each solution in the list
    for candidate_metric, soln in zip(all_pars_metrics, solutions):
        # Assume candidate is not dominated; check against all others
        if not any(all(other[0][i] <= candidate_metric[i] for i in range(len(candidate_metric))) and
                   any(other[0][i] < candidate_metric[i] for i in range(len(candidate_metric))) 
                   for other in pareto_front):
            # If no one in the current Pareto front dominates the candidate, add it
            pareto_front.append((candidate_metric, soln))
            # Remove any from the pareto_front that is dominated by the new candidate
            pareto_front = [front for front in pareto_front if not all(candidate_metric[i] <= front[0][i] for i in range(len(candidate_metric))) or not any(candidate_metric[i] < front[0][i] for i in range(len(candidate_metric)))]

    pareto_metrics = [front[0] for front in pareto_front]
    pareto_solutions = [front[1] for front in pareto_front]
    return pareto_metrics, pareto_solutions


import torch
import torch.optim.lr_scheduler as lr_scheduler
from metient.util import vertex_labeling_util as vutil
import numpy as np
import copy
from metient.lib import polytomy_resolver as prutil
from metient import metient as met
from metient.util.globals import *
import matplotlib.pyplot as plt
from tqdm import tqdm

PROGRESS_BAR = 0 # Keeps track of optimization progress using tqdm

class VertexLabelingSolver:
    def __init__(self, L, T, p, G, O, weights, config, num_sites, num_nodes_to_label,
                 node_idx_to_label, input_T, idx_to_observed_sites):
        self.L = L
        self.T = T
        self.p = p
        self.G = G
        self.O = O
        self.weights = weights
        self.config = config
        self.num_sites = num_sites
        self.num_nodes_to_label = num_nodes_to_label
        self.node_idx_to_label = node_idx_to_label
        self.input_T = input_T
        self.idx_to_observed_sites = idx_to_observed_sites
        # This gets set at first optimization time
        self.poly_res = None
        self.fixed_labeling = None
    
    def run(self):
        return run_multiple_optimizations(self)

def optimize_v(v_solver, X, poly_res, exploration_weights, max_iter, v_interval, is_second_optimization):
    
    # Unpack config
    lr = v_solver.config['lr']
    init_temp, final_temp = v_solver.config['init_temp'], v_solver.config['final_temp']
    t_anneal_rate, v_anneal_rate = v_solver.config['t_anneal_rate'], v_solver.config['v_anneal_rate']

    v_optimizer = torch.optim.Adam([X], lr=lr)
    v_scheduler = lr_scheduler.LinearLR(v_optimizer, start_factor=1.0, end_factor=0.5, total_iters=max_iter)
    
    solve_polytomies = v_solver.config['solve_polytomies']
    if solve_polytomies:
        poly_optimizer = torch.optim.Adam([poly_res.latent_var], lr=lr)
        poly_scheduler = lr_scheduler.LinearLR(poly_optimizer, start_factor=1.0, end_factor=0.5, total_iters=max_iter)
    v_temps, t_temps = [], []

    v_temp = init_temp
    t_temp = init_temp
    j = 0
    k = 0

    global PROGRESS_BAR
    for i in range(max_iter):
        update_path = update_path_matrix(i, max_iter, solve_polytomies, is_second_optimization)
        if solve_polytomies and update_path:
            poly_optimizer.zero_grad()

        v_optimizer.zero_grad()
        V, v_losses, soft_V, T = compute_v_loss(X, v_solver, poly_res, exploration_weights, update_path, v_temp, t_temp)
        mean_loss = torch.mean(v_losses)
        mean_loss.backward()

        if solve_polytomies and update_path:
            poly_optimizer.step()
            poly_scheduler.step()
            if i % 5 == 0:
                t_temp = np.maximum(t_temp * np.exp(-t_anneal_rate * j), final_temp)
            j += 1

        else:
            v_optimizer.step()
            v_scheduler.step()

            if i % v_interval == 0:
                v_temp = np.maximum(v_temp * np.exp(-v_anneal_rate * k), final_temp)
            k += 1

        v_temps.append(v_temp)
        t_temps.append(t_temp)

        PROGRESS_BAR.update(1)
    
    # plt.figure(figsize=(1,1),dpi=100)
    # plt.plot([x for x in range(len(v_temps))], v_temps, marker='.'); plt.show(); plt.close()
    # plt.figure(figsize=(1,1),dpi=100)
    # plt.plot([x for x in range(len(t_temps))], t_temps, marker='.'); plt.show(); plt.close()

    return V, soft_V, T, poly_res

def first_v_optimization(v_solver, exploration_weights):
    # We're learning X, which is the vertex labeling of the internal nodes
    X = x_weight_initialization(v_solver)
    X.requires_grad = True
    
    vutil.LAST_P = None
    
    solve_polytomies = v_solver.config['solve_polytomies']

    # If solving for polytomies, setup T and G appropriately
    if solve_polytomies:
        nodes_w_polys, resolver_sites = vutil.get_k_or_more_children_nodes(v_solver.input_T, v_solver.T, 
                                                                           v_solver.idx_to_observed_sites, 3, True, 2)
        # print("nodes_w_polys",nodes_w_polys, "resolver_sites", resolver_sites)
        if len(nodes_w_polys) == 0:
            print("No potential polytomies to solve, not resolving polytomies.")
            poly_res, solve_polytomies = None, False
        else:
            poly_res = prutil.PolytomyResolver(v_solver, nodes_w_polys, resolver_sites)
            #T, G, node_idx_to_label = poly_res.T, poly_res.G, poly_res.node_idx_to_label
    else:
        poly_res = None

    v_solver.poly_res = poly_res
    v_solver.config['solve_polytomies'] = solve_polytomies

    # First optimization
    optimized_Vs, _, optimized_Ts, _ = optimize_v(v_solver, X, v_solver.poly_res, exploration_weights, v_solver.config['first_max_iter'],
                                                  v_solver.config['first_v_interval'], False)
    
    # Identify optimal subtrees, keep them fixed, and solve for the rest of the tree
    optimal_nodes, optimal_batch_nums = find_optimal_subtrees(optimized_Ts, optimized_Vs, v_solver)

    return optimal_nodes, optimal_batch_nums, optimized_Ts, optimized_Vs

    
def second_v_optimization(v_solver, run_specific_x, run_specific_poly_res, exploration_weights):
    vutil.LAST_P = None

    # Second optimization
    optimzed_Vs, optimzed_soft_Vs, optimzed_Ts, run_specific_poly_res = optimize_v(v_solver, run_specific_x, run_specific_poly_res, exploration_weights,
                                                            v_solver.config['second_max_iter'], v_solver.config['second_v_interval'], True)


    return optimzed_Vs, optimzed_soft_Vs, optimzed_Ts, run_specific_poly_res

def full_exploration_weights(weights):
    return met.Weights(mig=DEFAULT_CALIBRATE_MIG_WEIGHTS, comig=DEFAULT_CALIBRATE_COMIG_WEIGHTS, 
                       seed_site=DEFAULT_CALIBRATE_SEED_WEIGHTS, data_fit=weights.data_fit, 
                       reg=weights.reg, entropy=weights.entropy, gen_dist=0.0, organotrop=0.0)


def run_multiple_optimizations(v_solver):
    global PROGRESS_BAR
    PROGRESS_BAR = tqdm(total=v_solver.config['first_max_iter'] + v_solver.config['second_max_iter']*v_solver.config['num_v_optimization_runs'], position=0)

    results = []
    
    # Only run first optimization once (this finds optimal subtrees)

    first_opt_result = first_v_optimization(v_solver, full_exploration_weights(v_solver.weights))
    optimal_nodes, optimal_batch_nums, optimized_Ts, optimized_Vs = first_opt_result

    # Function to wrap the second optimization 
    def second_optimization_task(v_solver, exploration_weights):
        # Each run needs its own polytomy resolver and X
        run_specific_poly_solver = copy.deepcopy(v_solver.poly_res)
        run_specific_x = x_weight_initialization(v_solver)
        run_specific_x, v_solver = initialize_optimal_x_polyres(run_specific_x, run_specific_poly_solver, optimal_nodes, optimal_batch_nums, optimized_Ts, optimized_Vs, v_solver)
        ret = second_v_optimization(v_solver, run_specific_x, run_specific_poly_solver, exploration_weights)
        return ret
    mig_weights = [1000.0,1.0,1.0]*100
    comig_weights = [1.0,1000.0,1.0]*100
    seed_weights = [1.0, 1.0,1000.0]*100
    order = [0,1,2]*100
    for r in range(v_solver.config['num_v_optimization_runs']):
        weight_idx = order[r]
        exploration_weights = met.Weights(mig=mig_weights[weight_idx], comig=comig_weights[weight_idx], 
                                         seed_site=seed_weights[weight_idx], data_fit=v_solver.weights.data_fit, 
                                         reg=v_solver.weights.reg, entropy=v_solver.weights.entropy, gen_dist=0.0, organotrop=0.0)
        
        # print(exploration_weights.mig,exploration_weights.comig, exploration_weights.seed_site)
        results.append(second_optimization_task(v_solver, exploration_weights))
    return results

def sample_gumbel(shape, eps=1e-8):
    G = torch.rand(shape)
    return -torch.log(-torch.log(G + eps) + eps)

def softmax_shifted_3d(X):
    shifted = X - X.max(dim=1, keepdim=True)[0]
    exps = torch.exp(shifted)
    return exps / exps.sum(dim=1, keepdim=True)

def gumbel_softmax_sample(logits, temperature):
    y = logits + sample_gumbel(logits.size())
    return softmax_shifted_3d(y / temperature)

def gumbel_softmax(logits, temperature, hard=True):
    '''
    Adapted from https://discuss.pytorch.org/t/stop-gradients-for-st-gumbel-softmax/530/5

    Sample from the Gumbel-Softmax distribution and optionally discretize.
    Args:
        logits: [batch_size, n_class] unnormalized log-probs
        temperature: non-negative scalar
        hard: if True, take argmax, but differentiate w.r.t. soft sample y
    Returns:
        [batch_size, n_class] sample from the Gumbel-Softmax distribution.
        If hard=True, then the returned sample will be one-hot, otherwise it will
        be a probabilitiy distribution that sums to 1 across classes

    '''
    shape = logits.size()
    assert len(shape) == 3 # [batch_size, num_sites, num_nodes]
    y_soft = gumbel_softmax_sample(logits, temperature)
    if hard:
        _, k = y_soft.max(1)
        y_hard = torch.zeros(shape, dtype=logits.dtype).scatter_(1, torch.unsqueeze(k, 1), 1.0)

        # This cool bit of code achieves two things:
        # (1) makes the output value exactly one-hot (since we add then subtract y_soft value)
        # (2) makes the gradient equal to y_soft gradient (since we strip all other gradients)
        y = y_hard - y_soft.detach() + y_soft
    else:
        y = y_soft
    return y, y_soft

def stack_vertex_labeling(L, X, p, poly_res, fixed_labeling):
    '''
    Use leaf labeling L and X (both of size batch_size x num_sites X num_internal_nodes)
    to get the anatomical sites of the leaf nodes and the internal nodes (respectively). 
    Stack the root labeling to get the full vertex labeling V. 
    '''
    # Expand leaf node labeling L to be repeated batch_size times
    bs = X.shape[0]
    L = vutil.repeat_n(L, bs)

    if fixed_labeling != None:
        full_X = torch.zeros((bs, X.shape[1], len(fixed_labeling.known_indices)+len(fixed_labeling.unknown_indices)))
        known_labelings = vutil.repeat_n(fixed_labeling.known_labelings, bs)
        full_X[:,:,fixed_labeling.unknown_indices] = X
        full_X[:,:,fixed_labeling.known_indices] = known_labelings
    else:
        full_X = X

    if poly_res != None:
        # Order is: internal nodes, new poly nodes, leaf nodes from U
        full_vert_labeling = torch.cat((full_X, vutil.repeat_n(poly_res.resolver_labeling, bs), L), dim=2)
    else:
        full_vert_labeling = torch.cat((full_X, L), dim=2)

    p = vutil.repeat_n(p, bs)
    # Concatenate the left part, new column, and right part along the second dimension
    return torch.cat((p, full_vert_labeling), dim=2)

def compute_v_loss(X, v_solver, poly_res, exploration_weights, update_path_matrix, v_temp, t_temp):
    '''
    Args:
        X: latent variable of labelings we are solving for. (batch_size x num_unknown_nodes x num_sites)
            where num_unkown_nodes = len(T) - (len(known_indices)), or len(unknown_indices)
        L: leaf node labels derived from U
        T: Full adjacency matrix which includes clone tree nodes as well as leaf nodes which were
            added from U > U_CUTOFF (observed in a site)
        p: one-hot vector indicating site of the primary
        G: Matrix of genetic distances between internal nodes (shape:  num_internal_nodes x num_internal_nodes).
        Lower values indicate lower branch lengths, i.e. more genetically similar.
        O: Array of frequencies with which the primary cancer type seeds site i (shape: num_anatomical_sites).  

    Returns:
        Loss of the labeling we're trying to learn (X) by computing maximum parsimony loss, organotropism, and
        genetic distance loss (if weights for genetic distance and organotropism != 0)
    '''
    softmax_X, softmax_X_soft = gumbel_softmax(X, v_temp)
    V = stack_vertex_labeling(v_solver.L, softmax_X, v_solver.p, v_solver.poly_res, v_solver.fixed_labeling)

    bs = X.shape[0]
    if poly_res != None:
        softmax_pol_res, _ = gumbel_softmax(poly_res.latent_var, t_temp)
        T = vutil.repeat_n(v_solver.T, bs)
        T[:,:,poly_res.children_of_polys] = softmax_pol_res
    else:
        T = vutil.repeat_n(v_solver.T, bs)

    G = v_solver.G
    if G != None:
        G = vutil.repeat_n(G, T.shape[0])
    loss, _ = vutil.clone_tree_labeling_objective(V, softmax_X_soft, T, v_solver.G, 
                                                  v_solver.O, v_solver.p, exploration_weights, update_path_matrix)
    return V, loss, softmax_X_soft, T

def x_weight_initialization(v_solver):

    nodes_w_children, biased_sites = vutil.get_k_or_more_children_nodes(v_solver.input_T, v_solver.T, v_solver.idx_to_observed_sites, 1, True, 1, cutoff=False)
    batch_size = v_solver.config['batch_size']
    # We're learning X, which is the vertex labeling of the internal nodes
    X = torch.rand(batch_size, v_solver.num_sites, v_solver.num_nodes_to_label)
    if v_solver.config['bias_weights']:
        eta = 3
        # Make 4 partitions: (1) biased towards the primary, (2) biased towards the primary + sites of children/grandchildren
        # (3) biased towards sites of children/grandchildren, (4) no bias
        quart = batch_size // 4
        
        # Bias for partitions [1-2]
        prim_site_idx = torch.nonzero(v_solver.p)[0][0]
        # This is really important to prevent large trees from getting stuck in local optima
        X[:quart*2,prim_site_idx,:] = eta / 2

        # Bias for partitions [2-3]
        # For each node, find which sites to bias labeling towards
        # by using the sites it and its children are detected in
        for node_idx, sites in zip(nodes_w_children, biased_sites):
            if node_idx == 0:
                continue # we know the root labeling
            idx = node_idx - 1
            for site_idx in sites:
                X[quart:quart*3,site_idx,idx] = eta

    return X

def update_path_matrix(itr, max_iter, solve_polytomies, second_optimization):
    if itr == -1:
        return True
    if not solve_polytomies:
        return False
    if second_optimization:
        return itr > max_iter*1/3 and itr < max_iter*2/3
    return itr > max_iter*1/3 and itr < max_iter*2/3

#################################################################
################### FIX OPTIMAL SUBTREES ########################
#################################################################

def find_optimal_subtree_nodes(optimized_Ts, optimized_Vs, num_internal_nodes):
    '''
    Args:
        - optimized_Ts: all possible solutions for possible adjacency matrices
        - optimized_Vs: all possible solutions for possible vertex labeling
    Returns:
        A list of node indices and their descendatns which belong to optimal subtrees (i.e.)
        all nodes in the subtree have the same color/label, and a list of the batch numbers
        that these optimal subtrees were found 
    '''

    P = vutil.path_matrix(optimized_Ts, remove_self_loops=False)
    VT = torch.transpose(optimized_Vs, 2, 1)
    # i,j = 1 if node i and node j have the same label 
    X = VT @ optimized_Vs
    same_color_subtrees = torch.logical_not(torch.logical_and(1 - X, P))
    # Get the indices of rows where all elements are True (all nodes have the same label)
    cand_optimal_subtree_indices = torch.nonzero(torch.all(same_color_subtrees, dim=2))
    # Tells us how many descendants each node has
    row_sums = torch.sum(P, dim=2)
    indexed_row_sums = torch.tensor([row_sums[idx[0]][idx[1]] for idx in cand_optimal_subtree_indices]).unsqueeze(1)
    cand_optimal_subtree_indices = torch.cat((cand_optimal_subtree_indices, indexed_row_sums), dim=1)
    # 2. Sort the optimal_subtrees by the number of children they have,
    # so that when we are solving for polytomies, we get the largest optimal subtrees possible
    cand_optimal_subtree_indices = cand_optimal_subtree_indices[cand_optimal_subtree_indices[:, 2].argsort(descending=True)]
    
    ## 3. Only keep optimal subtree roots that have a leaf node
    # nodes_w_leaves = vutil.nodes_w_leaf_nodes(optimized_Ts, num_internal_nodes)
    # cand_optimal_subtree_indices = [x for x in cand_optimal_subtree_indices if nodes_w_leaves[x[0], x[1]]]
    seen_nodes = set()
    optimal_batch_nums, optimal_subtree_nodes = [],[]
    for cand in cand_optimal_subtree_indices:
        batch_num = int(cand[0])
        optimal_subtree_root = int(cand[1])
        if optimal_subtree_root not in optimal_subtree_nodes:
            # Add the optimal_subtree_root and all its descendants
            descendants = [t.item() for t in torch.nonzero(P[batch_num,optimal_subtree_root])]
            
            # Don't fix clone's leaf nodes (num. descendants == 1), since we already know their labeling, 
            # and if they are under an optimal polytomy branch, they would be getting added by an optimal
            # subtree rooted by an ancestor
            if len(descendants) == 1: 
                continue
            # Don't fix a subtree where there are no leaf nodes 
            leaf_node_in_optimal_subtree = False
            for descendant in descendants:
                if descendant >= num_internal_nodes:
                    leaf_node_in_optimal_subtree = True
            
            if not leaf_node_in_optimal_subtree:
                continue
            # Don't fix when it's just a node and its leaf (the labeling of the node could be
            # multiple possibilities and still be optimal)
            if len(descendants) == 2:
                continue
            #print(optimal_subtree_root, descendants, leaf_node_in_optimal_subtree, batch_num)
            current_node_set = []
            for descendant in descendants:
                if descendant not in seen_nodes:
                    current_node_set.append(descendant)
                    seen_nodes.add(descendant)
            if len(current_node_set) > 0:
                optimal_batch_nums.append(batch_num)
                optimal_subtree_nodes.append(current_node_set)
    return optimal_subtree_nodes, optimal_batch_nums

def initialize_optimal_x_polyres(X, poly_res, optimal_subtree_nodes,optimal_batch_nums, optimized_Ts, optimized_Vs, v_solver):
    
    poly_resolver_to_optimal_children = {}
    known_indices = []
    known_labelings = []
    # TODO: re-initialize after first optimization?
    # if poly_res != None:
    #     # Anywhere the resolver is not -inf, reset the starting values to 1s
    #     poly_res.latent_var[poly_res.latent_var != float('-inf')] = 1

    # Fix node labels and node edges
    for optimal_subtree_set,optimal_batch_num in zip(optimal_subtree_nodes,optimal_batch_nums):
        for node_idx in optimal_subtree_set:
            # If this is a witness node from U or the root index, we already know its vertex labeling
            if node_idx <= X.shape[2] and node_idx != 0:
                optimal_site = int(optimized_Vs[optimal_batch_num,:,node_idx].nonzero(as_tuple=False))
                idx = node_idx - 1 # X doesn't include root node
                known_indices.append(idx)
                known_labelings.append(torch.eye(v_solver.num_sites)[optimal_site].T)
                X[:,optimal_site,idx] = 1
                non_optimal_sites = [i for i in range(v_solver.num_sites) if i != optimal_site]
                X[:,non_optimal_sites,idx] = float("-inf")

            # If this node is the child of a polytomy resolver node, fix its location
            # if the parent (the polytomy resolver node) belongs to the same optimal subtree
            if poly_res != None and node_idx in poly_res.children_of_polys:
                poly_idx = poly_res.children_of_polys.index(node_idx)
                parent_idx = int(optimized_Ts[optimal_batch_num,:,node_idx].nonzero(as_tuple=False))
                if parent_idx not in optimal_subtree_set:
                    continue
                poly_res.latent_var[:,parent_idx, poly_idx] = 1
                non_parents = [i for i in range(optimized_Ts.shape[1]) if i != parent_idx]
                poly_res.latent_var[:,non_parents,poly_idx] = float("-inf")
                poly_res.latent_var[:,non_parents,poly_idx] = float("-inf")

                # Don't let any other non-optimal children of this polytomy resolver move around
                if parent_idx in poly_res.resolver_indices:
                    if parent_idx not in poly_resolver_to_optimal_children:
                        optimal_children = vutil.get_child_indices(optimized_Ts[optimal_batch_num,:,:], [parent_idx])
                        poly_resolver_to_optimal_children[parent_idx] = optimal_children

    
    if poly_res != None:
        # Fix all other polytomy children s.t. they cannot move to be a child of the fixed node_idx
        # for parent_idx in poly_resolver_to_optimal_children:
            
        #     optimal_children = poly_resolver_to_optimal_children[parent_idx]
        #     optimal_children_poly_indices = [poly_res.children_of_polys.index(i) for i in optimal_children]
        #     other_children = [i for i in range(poly_res.latent_var.shape[2]) if i not in optimal_children_poly_indices]
        #     poly_res.latent_var[:,parent_idx,other_children] = float("-inf")
        poly_res.latent_var.requires_grad = True
    
    fixed_labeling = None
    if len(known_indices) != 0:
        unknown_indices = [x for x in range(v_solver.num_nodes_to_label) if x not in known_indices]
        known_labelings = torch.stack(known_labelings, dim=1)
        X = X[:,:,unknown_indices] # only include the unknown indices for inference
        fixed_labeling = vutil.FixedVertexLabeling(known_indices, unknown_indices, known_labelings)

    v_solver.fixed_labeling = fixed_labeling
    X.requires_grad = True

    return X, v_solver

def find_optimal_subtrees(optimized_Ts, optimized_Vs, v_solver):
    '''
    After the first round of optimization, there are optimal subtrees (subtrees where
    the labelings of *all* nodes is the same), which we can keep fixed, since there
    are no other more optimal labelings rooted at this branch.

    Two things we can fix: the labeling of the nodes in optimal subtrees,
    and the edges of the subtrees if polytomy resolution is being used. Search all
    samples to find optimal subtrees, since there might not be one solution with all 
    optimal subtrees.
    '''
    
    num_internal_nodes = v_solver.num_nodes_to_label + 1 # root node
    
    # Re-initialize optimized_Ts with the tree with the best subtree structure
    poly_res = v_solver.poly_res
    if poly_res != None:
        poly_res.latent_var.requires_grad = False
        num_internal_nodes += len(poly_res.resolver_indices)
        # print("poly_res.children_of_polys", poly_res.children_of_polys)
        # print("poly_res.resolver_indices", poly_res.resolver_indices)
        # print("poly_res.resolver_index_to_parent_idx",poly_res.resolver_index_to_parent_idx)
    
    # 1. Find samples with optimal subtrees
    optimal_subtree_nodes, optimal_batch_nums = find_optimal_subtree_nodes(optimized_Ts, optimized_Vs, num_internal_nodes)

    # # 3. Re-initialize X and polytomy resolver with optimal subtrees (labelings and structure) fixed. 
    # X = x_weight_initialization(v_solver)
    # X, v_solver = initalize_optimal_x_polyres(X, optimal_subtree_nodes, optimal_batch_nums, optimized_Ts, optimized_Vs, v_solver)
            
    return optimal_subtree_nodes, optimal_batch_nums


import torch
from metient.util import vertex_labeling_util as vutil
from torch.distributions.binomial import Binomial
import numpy as np

class ObservedClonesSolver:
    def __init__(self, num_sites, num_internal_nodes, ref, var, omega, idx_to_observed_sites,
                 B, input_T, G, node_idx_to_label, weights, config, estimate_observed_clones):
        self.ref = ref
        self.var = var
        self.omega = omega
        self.B = B
        self.input_T = input_T
        self.G = G
        self.weights = weights
        self.config = config
        self.num_sites = num_sites
        self.num_internal_nodes = num_internal_nodes
        self.node_idx_to_label = node_idx_to_label
        self.estimate_observed_clones = estimate_observed_clones
        self.idx_to_observed_sites = idx_to_observed_sites
    
    def run(self):
        if not self.estimate_observed_clones:
            T, G, L = vutil.full_adj_matrix_from_internal_node_idx_to_sites_present(self.input_T, self.G, self.idx_to_observed_sites, 
                                                                                    self.num_sites, self.config['identical_clone_gen_dist'])
            return None, self.input_T, T, G, L, self.node_idx_to_label, self.num_internal_nodes, self.idx_to_observed_sites

        return find_umap(self)

def find_umap(u_solver):
    
    # We're learning psi, which is the mixture matrix U (U = softmax(psi)), and tells us the existence
    # and anatomical locations of the extant clones (U > U_CUTOFF)
    #psi = -1 * torch.rand(num_sites, num_internal_nodes + 1) # an extra column for normal cells
    psi = torch.ones(u_solver.num_sites, u_solver.num_internal_nodes + 1) # an extra column for normal cells
    psi.requires_grad = True 
    u_optimizer = torch.optim.Adam([psi], lr=u_solver.config['lr'])

    i = 0
    u_prev = psi
    u_diff = 1e9
    losses = []
   
    while u_diff > 1e-6 and i < 300:
        u_optimizer.zero_grad()
        U, u_loss = compute_u_loss(psi, u_solver.ref, u_solver.var, u_solver.omega, u_solver.B, u_solver.weights)
        u_loss.backward()
        u_optimizer.step()
        u_diff = torch.abs(torch.norm(u_prev - U))
        u_prev = U
        i += 1
        losses.append(u_loss.detach().numpy())
    
    with torch.no_grad():
        full_T, full_G, L, idx_to_observed_sites = vutil.full_adj_matrix_using_inferred_observed_clones(U, u_solver.input_T, u_solver.G, u_solver.num_sites, u_solver.config['identical_clone_gen_dist'])

        # Remove any leaf nodes that aren't detected at > U_CUTOFF in any sites. These are not well estimated
        U_clones = U[:,1:]
        removal_indices = []
        for node_idx in range(U_clones.shape[1]):
            children = vutil.get_child_indices(u_solver.input_T, [node_idx])
            if node_idx not in idx_to_observed_sites and len(children) == 0:
                removal_indices.append(node_idx)
        #print("node indices not well estimated", removal_indices)
        U, input_T, T, G, node_idx_to_label, idx_to_observed_sites = vutil.remove_leaf_indices_not_observed_sites(removal_indices, U, u_solver.input_T, 
                                                                                                                  full_T, full_G, u_solver.node_idx_to_label, idx_to_observed_sites)
        num_internal_nodes = u_solver.num_internal_nodes - len(removal_indices)

    return U, input_T, T, G, L, node_idx_to_label, num_internal_nodes, idx_to_observed_sites

# Adapted from PairTree
def calc_llh(F_hat, R, V, omega_v):
    '''
    Args:
        F_hat: estimated subclonal frequency matrix (num_nodes x num_mutation_clusters)
        R: Reference allele count matrix (num_samples x num_mutation_clusters)
        V: Variant allele count matrix (num_samples x num_mutation_clusters)
    Returns:
        Data fit using the Binomial likelihood (p(x|F_hat)). See PairTree (Wintersinger et. al.)
        supplement section 2.2 for details.
    '''

    N = R + V
    S, K = F_hat.shape

    for matrix in V, N, omega_v:
        assert(matrix.shape == (S, K-1))

    P = torch.mul(omega_v, F_hat[:,1:])

    bin_dist = Binomial(N, P)
    F_llh = bin_dist.log_prob(V) / np.log(2)
    assert(not torch.any(F_llh.isnan()))
    assert(not torch.any(F_llh.isinf()))

    llh_per_sample = -torch.sum(F_llh, axis=1) / S
    nlglh = torch.sum(llh_per_sample) / (K-1)
    return nlglh

def compute_u_loss(psi, ref, var, omega, B, weights):
    '''
    Args:
        psi: raw values we are estimating of matrix U (num_sites x num_internal_nodes)
        ref: Reference matrix (num_anatomical_sites x num_mutation_clusters). Num. reads that map to reference allele
        var: Variant matrix (num_anatomical_sites x num_mutation_clusters). Num. reads that map to variant allele
        omega: VAF to subclonal frequency correction 
        B: Mutation matrix (shape: num_internal_nodes x num_mutation_clusters)
        weights: Weights object

    Returns:
        Loss to score the estimated proportions of each clone in each site
    '''
    
    # Using the softmax enforces that the row sums are 1, since the proprtions of
    # clones in a given site should sum to 1
    U = torch.softmax(psi, dim=1)
    # print("psi", psi)
    #print("U", U)

    # 1. Data fit
    F_hat = (U @ B)
    nlglh = calc_llh(F_hat, ref, var, omega)
    # 2. Regularization to make some values of U -> 0
    reg = torch.sum(psi) # l1 norm 
    # print("weights.reg", weights.reg, "reg", reg)
    # print("weights.data_fit", weights.data_fit, "nlglh", nlglh)
    clone_proportion_loss = (weights.data_fit*nlglh + weights.reg*reg)
    
    return U, clone_proportion_loss

def no_cna_omega(shape):
    '''
    Returns omega values assuming no copy number alterations (0.5)
    Shape is (num_anatomical_sites x num_mutation_clusters)
    '''
    return torch.ones(shape) * 0.5